/** @file function.cpp
 *
 *  Implementation of class of symbolic functions. */

/*
 *  This file was generated automatically by function.pl.
 *  Please do not modify it directly, edit the perl script instead!
 *  function.pl options: $maxargs=13
 *
 *  GiNaC Copyright (C) 1999-2003 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <string>
#include <stdexcept>
#include <list>

#include "function.h"
#include "fderivative.h"
#include "ex.h"
#include "lst.h"
#include "symmetry.h"
#include "print.h"
#include "archive.h"
#include "inifcns.h"
#include "tostring.h"
#include "utils.h"
#include "remember.h"

namespace GiNaC {

//////////
// helper class function_options
//////////

function_options::function_options()
{
	initialize();
}

function_options::function_options(std::string const & n, std::string const & tn)
{
	initialize();
	set_name(n,tn);
}

function_options::~function_options()
{
	// nothing to clean up at the moment
}

void function_options::initialize(void)
{
	set_name("unnamed_function","\\mbox{unnamed}");
	nparams = 0;
	eval_f = evalf_f = derivative_f = series_f = 0;
	evalf_params_first = true;
	use_return_type = false;
	eval_use_exvector_args = false;
	evalf_use_exvector_args = false;
	derivative_use_exvector_args = false;
	series_use_exvector_args = false;
	use_remember = false;
	functions_with_same_name = 1;
	symtree = 0;
}

function_options & function_options::set_name(std::string const & n,
                                              std::string const & tn)
{
	name = n;
	if (tn==std::string())
		TeX_name = "\\mbox{"+name+"}";
	else
		TeX_name = tn;
	return *this;
}

function_options & function_options::latex_name(std::string const & tn)
{
	TeX_name = tn;
	return *this;
}

// the following lines have been generated for max. 13 parameters
function_options & function_options::eval_func(eval_funcp_1 e)
{
	test_and_set_nparams(1);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_2 e)
{
	test_and_set_nparams(2);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_3 e)
{
	test_and_set_nparams(3);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_4 e)
{
	test_and_set_nparams(4);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_5 e)
{
	test_and_set_nparams(5);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_6 e)
{
	test_and_set_nparams(6);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_7 e)
{
	test_and_set_nparams(7);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_8 e)
{
	test_and_set_nparams(8);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_9 e)
{
	test_and_set_nparams(9);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_10 e)
{
	test_and_set_nparams(10);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_11 e)
{
	test_and_set_nparams(11);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_12 e)
{
	test_and_set_nparams(12);
	eval_f = eval_funcp(e);
	return *this;
}
function_options & function_options::eval_func(eval_funcp_13 e)
{
	test_and_set_nparams(13);
	eval_f = eval_funcp(e);
	return *this;
}

function_options & function_options::evalf_func(evalf_funcp_1 ef)
{
	test_and_set_nparams(1);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_2 ef)
{
	test_and_set_nparams(2);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_3 ef)
{
	test_and_set_nparams(3);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_4 ef)
{
	test_and_set_nparams(4);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_5 ef)
{
	test_and_set_nparams(5);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_6 ef)
{
	test_and_set_nparams(6);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_7 ef)
{
	test_and_set_nparams(7);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_8 ef)
{
	test_and_set_nparams(8);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_9 ef)
{
	test_and_set_nparams(9);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_10 ef)
{
	test_and_set_nparams(10);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_11 ef)
{
	test_and_set_nparams(11);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_12 ef)
{
	test_and_set_nparams(12);
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options & function_options::evalf_func(evalf_funcp_13 ef)
{
	test_and_set_nparams(13);
	evalf_f = evalf_funcp(ef);
	return *this;
}

function_options & function_options::derivative_func(derivative_funcp_1 d)
{
	test_and_set_nparams(1);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_2 d)
{
	test_and_set_nparams(2);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_3 d)
{
	test_and_set_nparams(3);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_4 d)
{
	test_and_set_nparams(4);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_5 d)
{
	test_and_set_nparams(5);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_6 d)
{
	test_and_set_nparams(6);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_7 d)
{
	test_and_set_nparams(7);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_8 d)
{
	test_and_set_nparams(8);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_9 d)
{
	test_and_set_nparams(9);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_10 d)
{
	test_and_set_nparams(10);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_11 d)
{
	test_and_set_nparams(11);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_12 d)
{
	test_and_set_nparams(12);
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options & function_options::derivative_func(derivative_funcp_13 d)
{
	test_and_set_nparams(13);
	derivative_f = derivative_funcp(d);
	return *this;
}

function_options & function_options::series_func(series_funcp_1 s)
{
	test_and_set_nparams(1);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_2 s)
{
	test_and_set_nparams(2);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_3 s)
{
	test_and_set_nparams(3);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_4 s)
{
	test_and_set_nparams(4);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_5 s)
{
	test_and_set_nparams(5);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_6 s)
{
	test_and_set_nparams(6);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_7 s)
{
	test_and_set_nparams(7);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_8 s)
{
	test_and_set_nparams(8);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_9 s)
{
	test_and_set_nparams(9);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_10 s)
{
	test_and_set_nparams(10);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_11 s)
{
	test_and_set_nparams(11);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_12 s)
{
	test_and_set_nparams(12);
	series_f = series_funcp(s);
	return *this;
}
function_options & function_options::series_func(series_funcp_13 s)
{
	test_and_set_nparams(13);
	series_f = series_funcp(s);
	return *this;
}

// end of generated lines

function_options& function_options::eval_func(eval_funcp_exvector e)
{
	eval_use_exvector_args = true;
	eval_f = eval_funcp(e);
	return *this;
}
function_options& function_options::evalf_func(evalf_funcp_exvector ef)
{
	evalf_use_exvector_args = true;
	evalf_f = evalf_funcp(ef);
	return *this;
}
function_options& function_options::derivative_func(derivative_funcp_exvector d)
{
	derivative_use_exvector_args = true;
	derivative_f = derivative_funcp(d);
	return *this;
}
function_options& function_options::series_func(series_funcp_exvector s)
{
	series_use_exvector_args = true;
	series_f = series_funcp(s);
	return *this;
}


function_options & function_options::set_return_type(unsigned rt, unsigned rtt)
{
	use_return_type = true;
	return_type = rt;
	return_type_tinfo = rtt;
	return *this;
}

function_options & function_options::do_not_evalf_params(void)
{
	evalf_params_first = false;
	return *this;
}

function_options & function_options::remember(unsigned size,
                                              unsigned assoc_size,
                                              unsigned strategy)
{
	use_remember = true;
	remember_size = size;
	remember_assoc_size = assoc_size;
	remember_strategy = strategy;
	return *this;
}

function_options & function_options::overloaded(unsigned o)
{
	functions_with_same_name = o;
	return *this;
}

function_options & function_options::set_symmetry(const symmetry & s)
{
	symtree = s;
	return *this;
}
	
void function_options::test_and_set_nparams(unsigned n)
{
	if (nparams==0) {
		nparams = n;
	} else if (nparams!=n) {
		// we do not throw an exception here because this code is
		// usually executed before main(), so the exception could not
		// caught anyhow
		std::cerr << "WARNING: number of parameters ("
		          << n << ") differs from number set before (" 
		          << nparams << ")" << std::endl;
	}
}

/** This can be used as a hook for external applications. */
unsigned function::current_serial = 0;


GINAC_IMPLEMENT_REGISTERED_CLASS(function, exprseq)

//////////
// default ctor, dtor, copy ctor, assignment operator and helpers
//////////

// public

function::function() : serial(0)
{
	tinfo_key = TINFO_function;
}

// protected

void function::copy(const function & other)
{
	inherited::copy(other);
	serial = other.serial;
}

void function::destroy(bool call_parent)
{
	if (call_parent)
		inherited::destroy(call_parent);
}

//////////
// other ctors
//////////

// public

function::function(unsigned ser) : serial(ser)
{
	tinfo_key = TINFO_function;
}

// the following lines have been generated for max. 13 parameters
function::function(unsigned ser, const ex & param1)
	: exprseq(param1), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2)
	: exprseq(param1, param2), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3)
	: exprseq(param1, param2, param3), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4)
	: exprseq(param1, param2, param3, param4), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5)
	: exprseq(param1, param2, param3, param4, param5), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6)
	: exprseq(param1, param2, param3, param4, param5, param6), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7)
	: exprseq(param1, param2, param3, param4, param5, param6, param7), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8, const ex & param9)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8, param9), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8, const ex & param9, const ex & param10)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8, param9, param10), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8, const ex & param9, const ex & param10, const ex & param11)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8, const ex & param9, const ex & param10, const ex & param11, const ex & param12)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12), serial(ser)
{
	tinfo_key = TINFO_function;
}
function::function(unsigned ser, const ex & param1, const ex & param2, const ex & param3, const ex & param4, const ex & param5, const ex & param6, const ex & param7, const ex & param8, const ex & param9, const ex & param10, const ex & param11, const ex & param12, const ex & param13)
	: exprseq(param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12, param13), serial(ser)
{
	tinfo_key = TINFO_function;
}

// end of generated lines

function::function(unsigned ser, const exprseq & es) : exprseq(es), serial(ser)
{
	tinfo_key = TINFO_function;
}

function::function(unsigned ser, const exvector & v, bool discardable) 
  : exprseq(v,discardable), serial(ser)
{
	tinfo_key = TINFO_function;
}

function::function(unsigned ser, exvector * vp) 
  : exprseq(vp), serial(ser)
{
	tinfo_key = TINFO_function;
}

//////////
// archiving
//////////

/** Construct object from archive_node. */
function::function(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
{
	// Find serial number by function name
	std::string s;
	if (n.find_string("name", s)) {
		unsigned int ser = 0;
		std::vector<function_options>::const_iterator i = registered_functions().begin(), iend = registered_functions().end();
		while (i != iend) {
			if (s == i->name) {
				serial = ser;
				return;
			}
			++i; ++ser;
		}
		throw (std::runtime_error("unknown function '" + s + "' in archive"));
	} else
		throw (std::runtime_error("unnamed function in archive"));
}

/** Unarchive the object. */
ex function::unarchive(const archive_node &n, const lst &sym_lst)
{
	return (new function(n, sym_lst))->setflag(status_flags::dynallocated);
}

/** Archive the object. */
void function::archive(archive_node &n) const
{
	inherited::archive(n);
	GINAC_ASSERT(serial < registered_functions().size());
	n.add_string("name", registered_functions()[serial].name);
}

//////////
// functions overriding virtual functions from base classes
//////////

// public

void function::print(const print_context & c, unsigned level) const
{
	GINAC_ASSERT(serial<registered_functions().size());

	if (is_of_type(c, print_tree)) {

		c.s << std::string(level, ' ') << class_name() << " "
		    << registered_functions()[serial].name
		    << std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
		    << ", nops=" << nops()
		    << std::endl;
		unsigned delta_indent = static_cast<const print_tree &>(c).delta_indent;
		for (unsigned i=0; i<seq.size(); ++i)
			seq[i].print(c, level + delta_indent);
		c.s << std::string(level + delta_indent, ' ') << "=====" << std::endl;

	} else if (is_of_type(c, print_csrc)) {

		// Print function name in lowercase
		std::string lname = registered_functions()[serial].name;
		unsigned num = lname.size();
		for (unsigned i=0; i<num; i++)
			lname[i] = tolower(lname[i]);
		c.s << lname << "(";

		// Print arguments, separated by commas
		exvector::const_iterator it = seq.begin(), itend = seq.end();
		while (it != itend) {
			it->print(c);
			++it;
			if (it != itend)
				c.s << ",";
		}
		c.s << ")";

	} else if (is_of_type(c, print_latex)) {
		c.s << registered_functions()[serial].TeX_name;
		printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
	} else {
		c.s << registered_functions()[serial].name;
		printseq(c, '(', ',', ')', exprseq::precedence(), function::precedence());
	}
}

ex function::expand(unsigned options) const
{
	// Only expand arguments when asked to do so
	if (options & expand_options::expand_function_args)
		return inherited::expand(options);
	else
		return (options == 0) ? setflag(status_flags::expanded) : *this;
}

ex function::eval(int level) const
{
	GINAC_ASSERT(serial<registered_functions().size());

	if (level>1) {
		// first evaluate children, then we will end up here again
		return function(serial,evalchildren(level));
	}

	const function_options &opt = registered_functions()[serial];

	// Canonicalize argument order according to the symmetry properties
	if (seq.size() > 1 && !(opt.symtree.is_zero())) {
		exvector v = seq;
		GINAC_ASSERT(is_a<symmetry>(opt.symtree));
		int sig = canonicalize(v.begin(), ex_to<symmetry>(opt.symtree));
		if (sig != INT_MAX) {
			// Something has changed while sorting arguments, more evaluations later
			if (sig == 0)
				return _ex0;
			return ex(sig) * thisexprseq(v);
		}
	}

	if (opt.eval_f==0) {
		return this->hold();
	}

	bool use_remember = opt.use_remember;
	ex eval_result;
	if (use_remember && lookup_remember_table(eval_result)) {
		return eval_result;
	}
	current_serial = serial;
	if (registered_functions()[serial].eval_use_exvector_args)
		eval_result = ((eval_funcp_exvector)(registered_functions()[serial].eval_f))(seq);
	else
	switch (opt.nparams) {
		// the following lines have been generated for max. 13 parameters
	case 1:
		eval_result = ((eval_funcp_1)(registered_functions()[serial].eval_f))(seq[1-1]);
		break;
	case 2:
		eval_result = ((eval_funcp_2)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1]);
		break;
	case 3:
		eval_result = ((eval_funcp_3)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1]);
		break;
	case 4:
		eval_result = ((eval_funcp_4)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1]);
		break;
	case 5:
		eval_result = ((eval_funcp_5)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1]);
		break;
	case 6:
		eval_result = ((eval_funcp_6)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1]);
		break;
	case 7:
		eval_result = ((eval_funcp_7)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1]);
		break;
	case 8:
		eval_result = ((eval_funcp_8)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1]);
		break;
	case 9:
		eval_result = ((eval_funcp_9)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1]);
		break;
	case 10:
		eval_result = ((eval_funcp_10)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1]);
		break;
	case 11:
		eval_result = ((eval_funcp_11)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1]);
		break;
	case 12:
		eval_result = ((eval_funcp_12)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1]);
		break;
	case 13:
		eval_result = ((eval_funcp_13)(registered_functions()[serial].eval_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1], seq[13-1]);
		break;

		// end of generated lines
	default:
		throw(std::logic_error("function::eval(): invalid nparams"));
	}
	if (use_remember) {
		store_remember_table(eval_result);
	}
	return eval_result;
}

ex function::evalf(int level) const
{
	GINAC_ASSERT(serial<registered_functions().size());

	// Evaluate children first
	exvector eseq;
	if (level == 1)
		eseq = seq;
	else if (level == -max_recursion_level)
		throw(std::runtime_error("max recursion level reached"));
	else
		eseq.reserve(seq.size());
	--level;
	exvector::const_iterator it = seq.begin(), itend = seq.end();
	while (it != itend) {
		eseq.push_back(it->evalf(level));
		++it;
	}
	
	if (registered_functions()[serial].evalf_f==0) {
		return function(serial,eseq).hold();
	}
	current_serial = serial;
	if (registered_functions()[serial].evalf_use_exvector_args)
		return ((evalf_funcp_exvector)(registered_functions()[serial].evalf_f))(seq);
	switch (registered_functions()[serial].nparams) {
		// the following lines have been generated for max. 13 parameters
	case 1:
		return ((evalf_funcp_1)(registered_functions()[serial].evalf_f))(eseq[1-1]);
	case 2:
		return ((evalf_funcp_2)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1]);
	case 3:
		return ((evalf_funcp_3)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1]);
	case 4:
		return ((evalf_funcp_4)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1]);
	case 5:
		return ((evalf_funcp_5)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1]);
	case 6:
		return ((evalf_funcp_6)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1]);
	case 7:
		return ((evalf_funcp_7)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1]);
	case 8:
		return ((evalf_funcp_8)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1]);
	case 9:
		return ((evalf_funcp_9)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1], eseq[9-1]);
	case 10:
		return ((evalf_funcp_10)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1], eseq[9-1], eseq[10-1]);
	case 11:
		return ((evalf_funcp_11)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1], eseq[9-1], eseq[10-1], eseq[11-1]);
	case 12:
		return ((evalf_funcp_12)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1], eseq[9-1], eseq[10-1], eseq[11-1], eseq[12-1]);
	case 13:
		return ((evalf_funcp_13)(registered_functions()[serial].evalf_f))(eseq[1-1], eseq[2-1], eseq[3-1], eseq[4-1], eseq[5-1], eseq[6-1], eseq[7-1], eseq[8-1], eseq[9-1], eseq[10-1], eseq[11-1], eseq[12-1], eseq[13-1]);

		// end of generated lines
	}
	throw(std::logic_error("function::evalf(): invalid nparams"));
}

unsigned function::calchash(void) const
{
	unsigned v = golden_ratio_hash(golden_ratio_hash(tinfo()) ^ serial);
	for (unsigned i=0; i<nops(); i++) {
		v = rotate_left_31(v);
		v ^= this->op(i).gethash();
	}
	v &= 0x7FFFFFFFU;
	if (flags & status_flags::evaluated) {
		setflag(status_flags::hash_calculated);
		hashvalue = v;
	}
	return v;
}

ex function::thisexprseq(const exvector & v) const
{
	return function(serial,v);
}

ex function::thisexprseq(exvector * vp) const
{
	return function(serial,vp);
}

/** Implementation of ex::series for functions.
 *  @see ex::series */
ex function::series(const relational & r, int order, unsigned options) const
{
	GINAC_ASSERT(serial<registered_functions().size());

	if (registered_functions()[serial].series_f==0) {
		return basic::series(r, order);
	}
	ex res;
	current_serial = serial;
	if (registered_functions()[serial].series_use_exvector_args) {
		try {
			res = ((series_funcp_exvector)(registered_functions()[serial].series_f))(seq, r, order, options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	}
	switch (registered_functions()[serial].nparams) {
		// the following lines have been generated for max. 13 parameters
	case 1:
		try {
			res = ((series_funcp_1)(registered_functions()[serial].series_f))(seq[1-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 2:
		try {
			res = ((series_funcp_2)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 3:
		try {
			res = ((series_funcp_3)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 4:
		try {
			res = ((series_funcp_4)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 5:
		try {
			res = ((series_funcp_5)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 6:
		try {
			res = ((series_funcp_6)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 7:
		try {
			res = ((series_funcp_7)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 8:
		try {
			res = ((series_funcp_8)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 9:
		try {
			res = ((series_funcp_9)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 10:
		try {
			res = ((series_funcp_10)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 11:
		try {
			res = ((series_funcp_11)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 12:
		try {
			res = ((series_funcp_12)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;
	case 13:
		try {
			res = ((series_funcp_13)(registered_functions()[serial].series_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1], seq[13-1],r,order,options);
		} catch (do_taylor) {
			res = basic::series(r, order, options);
		}
		return res;

		// end of generated lines
	}
	throw(std::logic_error("function::series(): invalid nparams"));
}

// protected

/** Implementation of ex::diff() for functions. It applies the chain rule,
 *  except for the Order term function.
 *  @see ex::diff */
ex function::derivative(const symbol & s) const
{
	ex result;

	if (serial == function_index_Order) {
		// Order Term function only differentiates the argument
		return Order(seq[0].diff(s));
	} else {
		// Chain rule
		ex arg_diff;
		unsigned num = seq.size();
		for (unsigned i=0; i<num; i++) {
			arg_diff = seq[i].diff(s);
			// We apply the chain rule only when it makes sense.  This is not
			// just for performance reasons but also to allow functions to
			// throw when differentiated with respect to one of its arguments
			// without running into trouble with our automatic full
			// differentiation:
			if (!arg_diff.is_zero())
				result += pderivative(i)*arg_diff;
		}
	}
	return result;
}

int function::compare_same_type(const basic & other) const
{
	GINAC_ASSERT(is_of_type(other, function));
	const function & o = static_cast<const function &>(other);

	if (serial != o.serial)
		return serial < o.serial ? -1 : 1;
	else
		return exprseq::compare_same_type(o);
}

bool function::is_equal_same_type(const basic & other) const
{
	GINAC_ASSERT(is_of_type(other, function));
	const function & o = static_cast<const function &>(other);

	if (serial != o.serial)
		return false;
	else
		return exprseq::is_equal_same_type(o);
}

bool function::match_same_type(const basic & other) const
{
	GINAC_ASSERT(is_of_type(other, function));
	const function & o = static_cast<const function &>(other);

	return serial == o.serial;
}

unsigned function::return_type(void) const
{
	if (seq.empty())
		return return_types::commutative;
	else
		return seq.begin()->return_type();
}

unsigned function::return_type_tinfo(void) const
{
	if (seq.empty())
		return tinfo_key;
	else
		return seq.begin()->return_type_tinfo();
}

//////////
// new virtual functions which can be overridden by derived classes
//////////

// none

//////////
// non-virtual functions in this class
//////////

// protected

ex function::pderivative(unsigned diff_param) const // partial differentiation
{
	GINAC_ASSERT(serial<registered_functions().size());
	
	// No derivative defined? Then return abstract derivative object
	if (registered_functions()[serial].derivative_f == NULL)
		return fderivative(serial, diff_param, seq);

	current_serial = serial;
	if (registered_functions()[serial].derivative_use_exvector_args)
		return ((derivative_funcp_exvector)(registered_functions()[serial].derivative_f))(seq, diff_param);
	switch (registered_functions()[serial].nparams) {
		// the following lines have been generated for max. 13 parameters
	case 1:
		return ((derivative_funcp_1)(registered_functions()[serial].derivative_f))(seq[1-1],diff_param);
	case 2:
		return ((derivative_funcp_2)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1],diff_param);
	case 3:
		return ((derivative_funcp_3)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1],diff_param);
	case 4:
		return ((derivative_funcp_4)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1],diff_param);
	case 5:
		return ((derivative_funcp_5)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1],diff_param);
	case 6:
		return ((derivative_funcp_6)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1],diff_param);
	case 7:
		return ((derivative_funcp_7)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1],diff_param);
	case 8:
		return ((derivative_funcp_8)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1],diff_param);
	case 9:
		return ((derivative_funcp_9)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1],diff_param);
	case 10:
		return ((derivative_funcp_10)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1],diff_param);
	case 11:
		return ((derivative_funcp_11)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1],diff_param);
	case 12:
		return ((derivative_funcp_12)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1],diff_param);
	case 13:
		return ((derivative_funcp_13)(registered_functions()[serial].derivative_f))(seq[1-1], seq[2-1], seq[3-1], seq[4-1], seq[5-1], seq[6-1], seq[7-1], seq[8-1], seq[9-1], seq[10-1], seq[11-1], seq[12-1], seq[13-1],diff_param);

		// end of generated lines
	}
	throw(std::logic_error("function::pderivative(): no diff function defined"));
}

std::vector<function_options> & function::registered_functions(void)
{
	static std::vector<function_options> * rf = new std::vector<function_options>;
	return *rf;
}

bool function::lookup_remember_table(ex & result) const
{
	return remember_table::remember_tables()[serial].lookup_entry(*this,result);
}

void function::store_remember_table(ex const & result) const
{
	remember_table::remember_tables()[serial].add_entry(*this,result);
}

// public

unsigned function::register_new(function_options const & opt)
{
	unsigned same_name = 0;
	for (unsigned i=0; i<registered_functions().size(); ++i) {
		if (registered_functions()[i].name==opt.name) {
			++same_name;
		}
	}
	if (same_name>=opt.functions_with_same_name) {
		// we do not throw an exception here because this code is
		// usually executed before main(), so the exception could not
		// caught anyhow
		std::cerr << "WARNING: function name " << opt.name
		          << " already in use!" << std::endl;
	}
	registered_functions().push_back(opt);
	if (opt.use_remember) {
		remember_table::remember_tables().
			push_back(remember_table(opt.remember_size,
			                         opt.remember_assoc_size,
			                         opt.remember_strategy));
	} else {
		remember_table::remember_tables().push_back(remember_table());
	}
	return registered_functions().size()-1;
}

/** Find serial number of function by name and number of parameters.
 *  Throws exception if function was not found. */
unsigned function::find_function(const std::string &name, unsigned nparams)
{
	std::vector<function_options>::const_iterator i = function::registered_functions().begin(), end = function::registered_functions().end();
	unsigned serial = 0;
	while (i != end) {
		if (i->get_name() == name && i->get_nparams() == nparams)
			return serial;
		++i;
		++serial;
	}
	throw (std::runtime_error("no function '" + name + "' with " + ToString(nparams) + " parameters defined"));
}

/** Return the print name of the function. */
std::string function::get_name(void) const
{
	GINAC_ASSERT(serial<registered_functions().size());
	return registered_functions()[serial].name;
}

} // namespace GiNaC

